from pynumaflow.mapper._dtypes import (
    Message,
    Messages,
    Datum,
    DROP,
)
from pynumaflow.mapper.async_server import AsyncMapper
from pynumaflow.mapper.multiproc_server import MultiProcMapper
from pynumaflow.mapper.server import Mapper

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "Mapper",
    "AsyncMapper",
    "MultiProcMapper",
]



from pynumaflow.reducer._dtypes import (
    Message,
    Messages,
    Datum,
    IntervalWindow,
    Metadata,
    DROP,
)
from pynumaflow.reducer.async_server import AsyncReducer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "IntervalWindow",
    "Metadata",
    "DROP",
    "AsyncReducer",
]



import logging
import multiprocessing
import os

from collections.abc import AsyncIterable

import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    MAP_STREAM_SOCK_PATH,
)
from pynumaflow.mapstreamer import Datum
from pynumaflow.mapstreamer._dtypes import MapStreamCallable
from pynumaflow.mapstreamer.proto import mapstream_pb2
from pynumaflow.mapstreamer.proto import mapstream_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow.info.server import get_sdk_version, write as info_server_write
from pynumaflow.info.types import ServerInfo, Protocol, Language, SERVER_INFO_FILE_PATH

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", 0)) or (_PROCESS_COUNT * 4)


class AsyncMapStreamer(mapstream_pb2_grpc.MapStreamServicer):
    """
    Provides an interface to write a Map Streamer
    which will be exposed over gRPC.

    Args:
        handler: Function callable following the type signature of MapStreamCallable
        sock_path: Path to the UNIX Domain Socket
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x4

    Example invocation:
    >>> from typing import Iterator
    >>> from pynumaflow.mapstreamer import Messages, Message \
    ...     Datum, AsyncMapStreamer
    ... import aiorun
    >>> async def map_stream_handler(key: [str], datums: Datum) -> AsyncIterable[Message]:
    ...    val = datum.value
    ...    _ = datum.event_time
    ...    _ = datum.watermark
    ...    for i in range(10):
    ...        yield Message(val, keys=keys)
    ...
    >>> grpc_server = AsyncMapStreamer(handler=map_stream_handler)
    >>> aiorun.run(grpc_server.start())
    """

    def __init__(
        self,
        handler: MapStreamCallable,
        sock_path=MAP_STREAM_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        self.__map_stream_handler: MapStreamCallable = handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    async def MapStreamFn(
        self,
        request: mapstream_pb2.MapStreamRequest,
        context: NumaflowServicerContext,
    ) -> AsyncIterable[mapstream_pb2.MapStreamResponse]:
        """
        Applies a map function to a datum stream in streaming mode.
        The pascal case function name comes from the proto mapstream_pb2_grpc.py file.
        """

        async for res in self.__invoke_map_stream(
            list(request.keys),
            Datum(
                keys=list(request.keys),
                value=request.value,
                event_time=request.event_time.ToDatetime(),
                watermark=request.watermark.ToDatetime(),
            ),
        ):
            yield mapstream_pb2.MapStreamResponse(result=res)

    async def __invoke_map_stream(self, keys: list[str], req: Datum):
        try:
            async for msg in self.__map_stream_handler(keys, req):
                yield mapstream_pb2.MapStreamResponse.Result(
                    keys=msg.keys, value=msg.value, tags=msg.tags
                )
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            raise err

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> mapstream_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto mapstream_pb2_grpc.py file.
        """
        return mapstream_pb2.ReadyResponse(ready=True)

    async def __serve_async(self, server) -> None:
        mapstream_pb2_grpc.add_MapStreamServicer_to_server(
            AsyncMapStreamer(handler=self.__map_stream_handler),
            server,
        )
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("GRPC Async Server listening on: %s", self.sock_path)
        await server.start()
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            version=get_sdk_version(),
        )
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)

        async def server_graceful_shutdown():
            """
            Shuts down the server with 5 seconds of grace period. During the
            grace period, the server won't accept new connections and allow
            existing RPCs to continue within the grace period.
            """
            _LOGGER.info("Starting graceful shutdown...")
            await server.stop(5)

        self.cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    async def start(self) -> None:
        """Starts the Async gRPC server on the given UNIX socket."""
        server = grpc.aio.server(options=self._server_options)
        await self.__serve_async(server)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: source.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0csource.proto\x12\tsource.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto"u\n\x0bReadRequest\x12/\n\x07request\x18\x01 \x01(\x0b\x32\x1e.source.v1.ReadRequest.Request\x1a\x35\n\x07Request\x12\x13\n\x0bnum_records\x18\x01 \x01(\x04\x12\x15\n\rtimeout_in_ms\x18\x02 \x01(\r"\xba\x01\n\x0cReadResponse\x12.\n\x06result\x18\x01 \x01(\x0b\x32\x1e.source.v1.ReadResponse.Result\x1az\n\x06Result\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12!\n\x06offset\x18\x02 \x01(\x0b\x32\x11.source.v1.Offset\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04keys\x18\x04 \x03(\t"k\n\nAckRequest\x12.\n\x07request\x18\x01 \x01(\x0b\x32\x1d.source.v1.AckRequest.Request\x1a-\n\x07Request\x12"\n\x07offsets\x18\x01 \x03(\x0b\x32\x11.source.v1.Offset"o\n\x0b\x41\x63kResponse\x12-\n\x06result\x18\x01 \x01(\x0b\x32\x1d.source.v1.AckResponse.Result\x1a\x31\n\x06Result\x12\'\n\x07success\x18\x01 \x01(\x0b\x32\x16.google.protobuf.Empty"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"]\n\x0fPendingResponse\x12\x31\n\x06result\x18\x01 \x01(\x0b\x32!.source.v1.PendingResponse.Result\x1a\x17\n\x06Result\x12\r\n\x05\x63ount\x18\x01 \x01(\x03".\n\x06Offset\x12\x0e\n\x06offset\x18\x01 \x01(\x0c\x12\x14\n\x0cpartition_id\x18\x02 \x01(\t2\xfb\x01\n\x06Source\x12;\n\x06ReadFn\x12\x16.source.v1.ReadRequest\x1a\x17.source.v1.ReadResponse0\x01\x12\x36\n\x05\x41\x63kFn\x12\x15.source.v1.AckRequest\x1a\x16.source.v1.AckResponse\x12?\n\tPendingFn\x12\x16.google.protobuf.Empty\x1a\x1a.source.v1.PendingResponse\x12;\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x18.source.v1.ReadyResponseb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "source_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_READREQUEST"]._serialized_start = 89
    _globals["_READREQUEST"]._serialized_end = 206
    _globals["_READREQUEST_REQUEST"]._serialized_start = 153
    _globals["_READREQUEST_REQUEST"]._serialized_end = 206
    _globals["_READRESPONSE"]._serialized_start = 209
    _globals["_READRESPONSE"]._serialized_end = 395
    _globals["_READRESPONSE_RESULT"]._serialized_start = 273
    _globals["_READRESPONSE_RESULT"]._serialized_end = 395
    _globals["_ACKREQUEST"]._serialized_start = 397
    _globals["_ACKREQUEST"]._serialized_end = 504
    _globals["_ACKREQUEST_REQUEST"]._serialized_start = 459
    _globals["_ACKREQUEST_REQUEST"]._serialized_end = 504
    _globals["_ACKRESPONSE"]._serialized_start = 506
    _globals["_ACKRESPONSE"]._serialized_end = 617
    _globals["_ACKRESPONSE_RESULT"]._serialized_start = 568
    _globals["_ACKRESPONSE_RESULT"]._serialized_end = 617
    _globals["_READYRESPONSE"]._serialized_start = 619
    _globals["_READYRESPONSE"]._serialized_end = 649
    _globals["_PENDINGRESPONSE"]._serialized_start = 651
    _globals["_PENDINGRESPONSE"]._serialized_end = 744
    _globals["_PENDINGRESPONSE_RESULT"]._serialized_start = 721
    _globals["_PENDINGRESPONSE_RESULT"]._serialized_end = 744
    _globals["_OFFSET"]._serialized_start = 746
    _globals["_OFFSET"]._serialized_end = 792
    _globals["_SOURCE"]._serialized_start = 795
    _globals["_SOURCE"]._serialized_end = 1046
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from typing import TypeVar

R = TypeVar("R", bound="Response")


@dataclass
class Response:
    """
    Class to define the important information for the event.
    Args:
        value: the payload of the event.
        no_broadcast: the flag to indicate whether the event should be broadcasted.
    >>> # Example usage
    >>> Response.broadcast_message(b"hello")
    >>> Response.no_broadcast_message()
    """

    __slots__ = ("value", "no_broadcast")

    value: bytes
    no_broadcast: bool

    @classmethod
    def broadcast_message(cls: type[R], value: bytes) -> R:
        """
        Returns a SideInputResponse object with the given value,
        and the No broadcast flag set to False. This event will be broadcasted.
        """
        return Response(value=value, no_broadcast=False)

    @classmethod
    def no_broadcast_message(cls: type[R]) -> R:
        """
        Returns a SideInputResponse object with the No broadcast flag set to True.
        This event will not be broadcasted.
        """
        return Response(value=b"", no_broadcast=True)



from pynumaflow.sourcer._dtypes import (
    Message,
    ReadRequest,
    PendingResponse,
    AckRequest,
    Offset,
)
from pynumaflow.sourcer.async_server import AsyncSourcer
from pynumaflow.sourcer.server import Sourcer

__all__ = [
    "Message",
    "ReadRequest",
    "PendingResponse",
    "AckRequest",
    "Offset",
    "AsyncSourcer",
    "Sourcer",
]



import logging
import multiprocessing
import os

from collections.abc import AsyncIterable
from google.protobuf import timestamp_pb2 as _timestamp_pb2
import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    SOURCE_SOCK_PATH,
)
from pynumaflow.sourcer import ReadRequest
from pynumaflow.sourcer._dtypes import AsyncSourceReadCallable, Offset, AckRequest
from pynumaflow.sourcer.proto import source_pb2
from pynumaflow.sourcer.proto import source_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow.info.server import get_sdk_version, write as info_server_write
from pynumaflow.info.types import ServerInfo, Protocol, Language, SERVER_INFO_FILE_PATH

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", "4"))


class AsyncSourcer(source_pb2_grpc.SourceServicer):
    """
    Provides an interface to write an Asynchronous Sourcer
    which will be exposed over gRPC.

    Args:
        read_handler: Function callable following the type signature of AsyncSourceReadCallable
        ack_handler: Function handler for AckFn
        pending_handler: Function handler for PendingFn
        sock_path: Path to the UNIX Domain Socket
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x4

    Example invocation:
    >>> from typing import Iterator
    >>> from pynumaflow.sourcer import Message \
    ...     ReadRequest, AsyncSourcer
    ... import aiorun
    ... async def read_handler(datum: ReadRequest) -> AsyncIterable[Message]:
    ...     payload = b"payload:test_mock_message"
    ...     keys = ["test_key"]
    ...     offset = mock_offset()
    ...     event_time = mock_event_time()
    ...     for i in range(10):
    ...         yield Message(payload=payload, keys=keys, offset=offset, event_time=event_time)
    ... async def ack_handler(ack_request: AckRequest):
    ...     return
    ... async def pending_handler() -> PendingResponse:
    ...     PendingResponse(count=10)
    >>> grpc_server = AsyncSourcer(read_handler=read_handler,
    ...                     ack_handler=ack_handler,
    ...                     pending_handler=pending_handler)
    >>> aiorun.run(grpc_server.start())
    """

    def __init__(
        self,
        read_handler: AsyncSourceReadCallable,
        ack_handler,
        pending_handler,
        sock_path=SOURCE_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        if read_handler is None or ack_handler is None or pending_handler is None:
            raise ValueError("read_handler, ack_handler and pending_handler are required")
        self.__source_read_handler: AsyncSourceReadCallable = read_handler
        self.__source_ack_handler = ack_handler
        self.__source_pending_handler = pending_handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    async def ReadFn(
        self,
        request: source_pb2.ReadRequest,
        context: NumaflowServicerContext,
    ) -> AsyncIterable[source_pb2.ReadResponse]:
        """
        Applies a Read function and returns a stream of datum responses.
        The pascal case function name comes from the proto source_pb2_grpc.py file.
        """

        async for res in self.__invoke_source_read_stream(
            ReadRequest(
                num_records=request.request.num_records,
                timeout_in_ms=request.request.timeout_in_ms,
            )
        ):
            yield source_pb2.ReadResponse(result=res)

    async def __invoke_source_read_stream(self, req: ReadRequest):
        try:
            async for msg in self.__source_read_handler(req):
                event_time_timestamp = _timestamp_pb2.Timestamp()
                event_time_timestamp.FromDatetime(dt=msg.event_time)
                yield source_pb2.ReadResponse.Result(
                    payload=msg.payload,
                    keys=msg.keys,
                    offset=msg.offset.as_dict,
                    event_time=event_time_timestamp,
                )
        except Exception as err:
            _LOGGER.critical("User-Defined Source ReadError ", exc_info=True)
            raise err

    async def AckFn(
        self, request: source_pb2.AckRequest, context: NumaflowServicerContext
    ) -> source_pb2.AckResponse:
        """
        Applies an Ack function in User Defined Source
        """
        # proto repeated field(offsets) is of type google._upb._message.RepeatedScalarContainer
        # we need to explicitly convert it to list
        offsets = []
        for offset in request.request.offsets:
            offsets.append(Offset(offset.offset, offset.partition_id))
        try:
            await self.__invoke_ack(ack_req=request)
        except Exception as e:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(str(e))
            raise e

        return source_pb2.AckResponse()

    async def __invoke_ack(self, ack_req: AckRequest):
        """
        Invokes the Source Ack Function.
        """
        try:
            await self.__source_ack_handler(ack_req)
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            raise err
        return source_pb2.AckResponse.Result()

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> source_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto source_pb2_grpc.py file.
        """
        return source_pb2.ReadyResponse(ready=True)

    async def PendingFn(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> source_pb2.PendingResponse:
        """
        PendingFn returns the number of pending records
        at the user defined source.
        """
        try:
            count = await self.__source_pending_handler()
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            raise err
        resp = source_pb2.PendingResponse.Result(count=count.count)
        return source_pb2.PendingResponse(result=resp)

    async def __serve_async(self, server) -> None:
        source_pb2_grpc.add_sourceServicer_to_server(
            AsyncSourcer(read_handler=self.__source_read_handler),
            server,
        )
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("GRPC Async Server listening on: %s", self.sock_path)
        await server.start()
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            version=get_sdk_version(),
        )
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)

        async def server_graceful_shutdown():
            """
            Shuts down the server with 5 seconds of grace period. During the
            grace period, the server won't accept new connections and allow
            existing RPCs to continue within the grace period.
            """
            _LOGGER.info("Starting graceful shutdown...")
            await server.stop(5)

        self.cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    async def start(self) -> None:
        """Starts the Async gRPC server on the given UNIX socket."""
        server = grpc.aio.server(options=self._server_options)
        await self.__serve_async(server)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import source_pb2 as source__pb2


class SourceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ReadFn = channel.unary_stream(
            "/source.v1.Source/ReadFn",
            request_serializer=source__pb2.ReadRequest.SerializeToString,
            response_deserializer=source__pb2.ReadResponse.FromString,
        )
        self.AckFn = channel.unary_unary(
            "/source.v1.Source/AckFn",
            request_serializer=source__pb2.AckRequest.SerializeToString,
            response_deserializer=source__pb2.AckResponse.FromString,
        )
        self.PendingFn = channel.unary_unary(
            "/source.v1.Source/PendingFn",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=source__pb2.PendingResponse.FromString,
        )
        self.IsReady = channel.unary_unary(
            "/source.v1.Source/IsReady",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=source__pb2.ReadyResponse.FromString,
        )


class SourceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ReadFn(self, request, context):
        """Read returns a stream of datum responses.
        The size of the returned ReadResponse is less than or equal to the num_records specified in ReadRequest.
        If the request timeout is reached on server side, the returned ReadResponse will contain all the datum that have been read (which could be an empty list).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AckFn(self, request, context):
        """AckFn acknowledges a list of datum offsets.
        When AckFn is called, it implicitly indicates that the datum stream has been processed by the source vertex.
        The caller (numa) expects the AckFn to be successful, and it does not expect any errors.
        If there are some irrecoverable errors when the callee (UDSource) is processing the AckFn request,
        then it is best to crash because there are no other retry mechanisms possible.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def PendingFn(self, request, context):
        """PendingFn returns the number of pending records at the user defined source."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(self, request, context):
        """IsReady is the heartbeat endpoint for user defined source gRPC."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_SourceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ReadFn": grpc.unary_stream_rpc_method_handler(
            servicer.ReadFn,
            request_deserializer=source__pb2.ReadRequest.FromString,
            response_serializer=source__pb2.ReadResponse.SerializeToString,
        ),
        "AckFn": grpc.unary_unary_rpc_method_handler(
            servicer.AckFn,
            request_deserializer=source__pb2.AckRequest.FromString,
            response_serializer=source__pb2.AckResponse.SerializeToString,
        ),
        "PendingFn": grpc.unary_unary_rpc_method_handler(
            servicer.PendingFn,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=source__pb2.PendingResponse.SerializeToString,
        ),
        "IsReady": grpc.unary_unary_rpc_method_handler(
            servicer.IsReady,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=source__pb2.ReadyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("source.v1.Source", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Source(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ReadFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/source.v1.Source/ReadFn",
            source__pb2.ReadRequest.SerializeToString,
            source__pb2.ReadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AckFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/source.v1.Source/AckFn",
            source__pb2.AckRequest.SerializeToString,
            source__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def PendingFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/source.v1.Source/PendingFn",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            source__pb2.PendingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def IsReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/source.v1.Source/IsReady",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            source__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from pynumaflow.sourcetransformer._dtypes import Message, Messages, Datum, DROP, EVENT_TIME_FOR_DROP
from pynumaflow.sourcetransformer.multiproc_server import MultiProcSourceTransformer
from pynumaflow.sourcetransformer.server import SourceTransformer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "EVENT_TIME_FOR_DROP",
    "SourceTransformer",
    "MultiProcSourceTransformer",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sideinput.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0fsideinput.proto\x12\x0csideinput.v1\x1a\x1bgoogle/protobuf/empty.proto"8\n\x11SideInputResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x12\x14\n\x0cno_broadcast\x18\x02 \x01(\x08"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\x99\x01\n\tSideInput\x12L\n\x11RetrieveSideInput\x12\x16.google.protobuf.Empty\x1a\x1f.sideinput.v1.SideInputResponse\x12>\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x1b.sideinput.v1.ReadyResponseb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "sideinput_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_SIDEINPUTRESPONSE"]._serialized_start = 62
    _globals["_SIDEINPUTRESPONSE"]._serialized_end = 118
    _globals["_READYRESPONSE"]._serialized_start = 120
    _globals["_READYRESPONSE"]._serialized_end = 150
    _globals["_SIDEINPUT"]._serialized_start = 153
    _globals["_SIDEINPUT"]._serialized_end = 306
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import transform_pb2 as transform__pb2


class SourceTransformStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SourceTransformFn = channel.unary_unary(
            "/sourcetransformer.v1.SourceTransform/SourceTransformFn",
            request_serializer=transform__pb2.SourceTransformRequest.SerializeToString,
            response_deserializer=transform__pb2.SourceTransformResponse.FromString,
        )
        self.IsReady = channel.unary_unary(
            "/sourcetransformer.v1.SourceTransform/IsReady",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=transform__pb2.ReadyResponse.FromString,
        )


class SourceTransformServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SourceTransformFn(self, request, context):
        """SourceTransformFn applies a function to each request element.
        In addition to map function, SourceTransformFn also supports assigning a new event time to response.
        SourceTransformFn can be used only at source vertex by source data transformer.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(self, request, context):
        """IsReady is the heartbeat endpoint for gRPC."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_SourceTransformServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "SourceTransformFn": grpc.unary_unary_rpc_method_handler(
            servicer.SourceTransformFn,
            request_deserializer=transform__pb2.SourceTransformRequest.FromString,
            response_serializer=transform__pb2.SourceTransformResponse.SerializeToString,
        ),
        "IsReady": grpc.unary_unary_rpc_method_handler(
            servicer.IsReady,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=transform__pb2.ReadyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "sourcetransformer.v1.SourceTransform", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class SourceTransform(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SourceTransformFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/sourcetransformer.v1.SourceTransform/SourceTransformFn",
            transform__pb2.SourceTransformRequest.SerializeToString,
            transform__pb2.SourceTransformResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def IsReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/sourcetransformer.v1.SourceTransform/IsReady",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            transform__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import logging
import multiprocessing
import os
from collections.abc import AsyncIterable

import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    SINK_SOCK_PATH,
    MAX_MESSAGE_SIZE,
)
from pynumaflow.info.server import get_sdk_version, write as info_server_write
from pynumaflow.info.types import ServerInfo, Protocol, Language, SERVER_INFO_FILE_PATH
from pynumaflow.sinker import Responses, Datum, Response
from pynumaflow.sinker._dtypes import SinkCallable
from pynumaflow.sinker.proto import sink_pb2_grpc, sink_pb2
from pynumaflow.types import NumaflowServicerContext

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", 0)) or (_PROCESS_COUNT * 4)


async def datum_generator(
    request_iterator: AsyncIterable[sink_pb2.SinkRequest],
) -> AsyncIterable[Datum]:
    async for d in request_iterator:
        datum = Datum(
            keys=list(d.keys),
            sink_msg_id=d.id,
            value=d.value,
            event_time=d.event_time.ToDatetime(),
            watermark=d.watermark.ToDatetime(),
        )
        yield datum


class AsyncSinker(sink_pb2_grpc.SinkServicer):
    """
    Provides an interface to write an Async Sinker
    which will be exposed over an Asyncronous gRPC server.

    Args:
        handler: Function callable following the type signature of SinkCallable
        sock_path: Path to the UNIX Domain Socket
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x 4

    Example invocation:
    >>> import aiorun
    >>> from pynumaflow.sinker import Datum, Responses, Response, AsyncSinker
    >>> async def my_handler(datums: AsyncIterable[Datum]) -> Responses:
    ...   responses = Responses()
    ...   async for msg in datums:
    ...     responses.append(Response.as_success(msg.id))
    ...   return responses
    >>> grpc_server = AsyncSinker(handler=my_handler)
    >>> aiorun.run(grpc_server.start())
    """

    def __init__(
        self,
        handler: SinkCallable,
        sock_path=SINK_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        self.background_tasks = set()
        self.__sink_handler: SinkCallable = handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    async def SinkFn(
        self,
        request_iterator: AsyncIterable[sink_pb2.SinkRequest],
        context: NumaflowServicerContext,
    ) -> sink_pb2.SinkResponse:
        """
        Applies a sink function to a list of datum elements.
        The pascal case function name comes from the proto sink_pb2_grpc.py file.
        """
        # if there is an exception, we will mark all the responses as a failure
        datum_iterator = datum_generator(request_iterator=request_iterator)
        results = await self.__invoke_sink(datum_iterator)

        return sink_pb2.SinkResponse(results=results)

    async def __invoke_sink(self, datum_iterator: AsyncIterable[Datum]):
        try:
            rspns = await self.__sink_handler(datum_iterator)
        except Exception as err:
            err_msg = "UDSinkError: %r" % err
            _LOGGER.critical(err_msg, exc_info=True)
            rspns = Responses()
            async for _datum in datum_iterator:
                rspns.append(Response.as_failure(_datum.id, err_msg))
        responses = []
        for rspn in rspns:
            responses.append(
                sink_pb2.SinkResponse.Result(id=rspn.id, success=rspn.success, err_msg=rspn.err)
            )
        return responses

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sink_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto sink_pb2_grpc.py file.
        """
        return sink_pb2.ReadyResponse(ready=True)

    async def __serve_async(self, server) -> None:
        sink_pb2_grpc.add_SinkServicer_to_server(AsyncSinker(self.__sink_handler), server)
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("GRPC Async Server listening on: %s", self.sock_path)
        await server.start()
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            version=get_sdk_version(),
        )
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)

        async def server_graceful_shutdown():
            _LOGGER.info("Starting graceful shutdown...")
            """
            Shuts down the server with 5 seconds of grace period. During the
            grace period, the server won't accept new connections and allow
            existing RPCs to continue within the grace period.
            await server.stop(5)
            """

        self.cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    async def start(self) -> None:
        """Starts the Async gRPC server on the given UNIX socket."""
        server = grpc.aio.server(options=self._server_options)
        await self.__serve_async(server)



from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar, Callable
from warnings import warn

from pynumaflow._constants import DROP, EVENT_TIME_FOR_DROP

M = TypeVar("M", bound="Message")
Ms = TypeVar("Ms", bound="Messages")


@dataclass(init=False)
class Message:
    """
    Basic datatype for data passing to the next vertex/vertices.

    Args:
        value: data in bytes
        event_time: event time of the message, usually extracted from the payload.
        keys: []string keys for vertex (optional)
        tags: []string tags for conditional forwarding (optional)
    """

    __slots__ = ("_value", "_keys", "_tags", "_event_time")

    _keys: list[str]
    _tags: list[str]
    _value: bytes
    _event_time: datetime

    def __init__(
        self, value: bytes, event_time: datetime, keys: list[str] = None, tags: list[str] = None
    ):
        """
        Creates a Message object to send value to a vertex.
        """
        self._tags = tags or []
        self._keys = keys or []

        # There is no year 0, so setting following as default event time.
        self._event_time = event_time or datetime(1, 1, 1, 0, 0)
        self._value = value or b""

    @classmethod
    def to_drop(cls: type[M]) -> M:
        return cls(b"", EVENT_TIME_FOR_DROP, None, [DROP])

    @property
    def event_time(self) -> datetime:
        return self._event_time

    @property
    def keys(self) -> list[str]:
        return self._keys

    @property
    def value(self) -> bytes:
        return self._value

    @property
    def tags(self) -> list[str]:
        return self._tags


class Messages(Sequence[M]):
    """
    Class to define a list of Message objects.

    Args:
        messages: list of Message objects.
    """

    __slots__ = ("_messages",)

    def __init__(self, *messages: M):
        self._messages = list(messages) or []

    def __str__(self):
        return str(self._messages)

    def __repr__(self):
        return str(self)

    def __len__(self) -> int:
        return len(self._messages)

    def __iter__(self) -> Iterator[M]:
        return iter(self._messages)

    def __getitem__(self, index: int) -> M:
        if isinstance(index, slice):
            raise TypeError("Slicing is not supported for Messages")
        return self._messages[index]

    def append(self, message: Message) -> None:
        self._messages.append(message)

    def items(self) -> list[Message]:
        warn(
            "Using items is deprecated and will be removed in v0.5. "
            "Iterate or index the Messages object instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._messages


@dataclass(init=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.sourcetransformer import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> d = Datum(
    ...       keys=["test_key"],
    ...       value=payload,
    ...       event_time=t1,
    ...       watermark=t2,
    ...    )
    """

    __slots__ = ("_keys", "_value", "_event_time", "_watermark")

    _keys: list[str]
    _value: bytes
    _event_time: datetime
    _watermark: datetime

    def __init__(
        self,
        keys: list[str],
        value: bytes,
        event_time: datetime,
        watermark: datetime,
    ):
        self._keys = keys or list()
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark

    def keys(self) -> list[str]:
        """Returns the keys of the event"""
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark


SourceTransformCallable = Callable[[list[str], Datum], Messages]



import asyncio
import logging
import multiprocessing
import os

from datetime import datetime, timezone
from collections.abc import AsyncIterable

import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    WIN_START_TIME,
    WIN_END_TIME,
    MAX_MESSAGE_SIZE,
    STREAM_EOF,
    DELIMITER,
    REDUCE_SOCK_PATH,
)
from pynumaflow.reducer import Datum, IntervalWindow, Metadata
from pynumaflow.reducer._dtypes import ReduceResult, ReduceCallable
from pynumaflow.reducer.asynciter import NonBlockingIterator
from pynumaflow.reducer.proto import reduce_pb2
from pynumaflow.reducer.proto import reduce_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow.info.server import get_sdk_version, write as info_server_write
from pynumaflow.info.types import ServerInfo, Protocol, Language, SERVER_INFO_FILE_PATH

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", 0)) or (_PROCESS_COUNT * 4)


async def datum_generator(
    request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
) -> AsyncIterable[Datum]:
    async for d in request_iterator:
        datum = Datum(
            keys=list(d.keys),
            value=d.value,
            event_time=d.event_time.ToDatetime(),
            watermark=d.watermark.ToDatetime(),
        )
        yield datum


class AsyncReducer(reduce_pb2_grpc.ReduceServicer):
    """
    Provides an interface to write a Reduce Function
    which will be exposed over gRPC.

    Args:
        handler: Function callable following the type signature of ReduceCallable
        sock_path: Path to the UNIX Domain Socket
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x4

    Example invocation:
    >>> from typing import Iterator
    >>> from pynumaflow.reducer import Messages, Message\
    ...     Datum, Metadata, AsyncReducer
    ... import aiorun
    ...
    >>> async def reduce_handler(key: list[str], datums: AsyncIterable[Datum],
    >>> md: Metadata) -> Messages:
    ...   interval_window = md.interval_window
    ...   counter = 0
    ...   async for _ in datums:
    ...     counter += 1
    ...   msg = (
    ...       f"counter:{counter} interval_window_start:{interval_window.start} "
    ...       f"interval_window_end:{interval_window.end}"
    ...   )
    ...   return Messages(Message(value=str.encode(msg), keys=keys))
    ...
    >>> grpc_server = AsyncReducer(handler=reduce_handler)
    >>> aiorun.run(grpc_server.start())
    """

    def __init__(
        self,
        handler: ReduceCallable,
        sock_path=REDUCE_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        self.__reduce_handler: ReduceCallable = handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    async def ReduceFn(
        self,
        request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
        context: NumaflowServicerContext,
    ) -> reduce_pb2.ReduceResponse:
        """
        Applies a reduce function to a datum stream.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.
        """

        start, end = None, None
        for metadata_key, metadata_value in context.invocation_metadata():
            if metadata_key == WIN_START_TIME:
                start = metadata_value
            elif metadata_key == WIN_END_TIME:
                end = metadata_value
        if not (start or end):
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            context.set_details(
                f"Expected to have all key/window_start_time/window_end_time; "
                f"got start: {start}, end: {end}."
            )
            yield reduce_pb2.ReduceResponse(results=[])
            return

        start_dt = datetime.fromtimestamp(int(start) / 1e3, timezone.utc)
        end_dt = datetime.fromtimestamp(int(end) / 1e3, timezone.utc)
        interval_window = IntervalWindow(start=start_dt, end=end_dt)

        datum_iterator = datum_generator(request_iterator=request_iterator)

        response_task = asyncio.create_task(
            self.__async_reduce_handler(interval_window, datum_iterator)
        )

        # Save a reference to the result of this function, to avoid a
        # task disappearing mid-execution.
        self.background_tasks.add(response_task)
        response_task.add_done_callback(lambda t: self.background_tasks.remove(t))

        await response_task
        results_futures = response_task.result()

        try:
            for fut in results_futures:
                await fut
                yield reduce_pb2.ReduceResponse(results=fut.result())
        except Exception as e:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(e.__str__())
            yield reduce_pb2.ReduceResponse(results=[])

    async def __async_reduce_handler(self, interval_window, datum_iterator: AsyncIterable[Datum]):
        callable_dict = {}
        # iterate through all the values
        async for d in datum_iterator:
            keys = d.keys()
            unified_key = DELIMITER.join(keys)
            result = callable_dict.get(unified_key, None)

            if not result:
                niter = NonBlockingIterator()
                riter = niter.read_iterator()
                # schedule an async task for consumer
                # returns a future that will give the results later.
                task = asyncio.create_task(
                    self.__invoke_reduce(keys, riter, Metadata(interval_window=interval_window))
                )
                # Save a reference to the result of this function, to avoid a
                # task disappearing mid-execution.
                self.background_tasks.add(task)
                task.add_done_callback(lambda t: self.background_tasks.remove(t))
                result = ReduceResult(task, niter, keys)

                callable_dict[unified_key] = result

            await result.iterator.put(d)

        for unified_key in callable_dict:
            await callable_dict[unified_key].iterator.put(STREAM_EOF)

        tasks = []
        for unified_key in callable_dict:
            fut = callable_dict[unified_key].future
            tasks.append(fut)

        return tasks

    async def __invoke_reduce(
        self, keys: list[str], request_iterator: AsyncIterable[Datum], md: Metadata
    ):
        try:
            msgs = await self.__reduce_handler(keys, request_iterator, md)
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            raise err

        datum_responses = []
        for msg in msgs:
            datum_responses.append(
                reduce_pb2.ReduceResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags)
            )

        return datum_responses

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> reduce_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.
        """
        return reduce_pb2.ReadyResponse(ready=True)

    async def __serve_async(self, server) -> None:
        reduce_pb2_grpc.add_ReduceServicer_to_server(
            AsyncReducer(handler=self.__reduce_handler),
            server,
        )
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("GRPC Async Server listening on: %s", self.sock_path)
        await server.start()
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            version=get_sdk_version(),
        )
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)

        async def server_graceful_shutdown():
            """
            Shuts down the server with 5 seconds of grace period. During the
            grace period, the server won't accept new connections and allow
            existing RPCs to continue within the grace period.
            """
            _LOGGER.info("Starting graceful shutdown...")
            await server.stop(5)

        self.cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    async def start(self) -> None:
        """Starts the Async gRPC server on the given UNIX socket."""
        server = grpc.aio.server(options=self._server_options)
        await self.__serve_async(server)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sink.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nsink.proto\x12\x07sink.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x95\x01\n\x0bSinkRequest\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\n\n\x02id\x18\x05 \x01(\t"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"u\n\x0cSinkResponse\x12-\n\x07results\x18\x01 \x03(\x0b\x32\x1c.sink.v1.SinkResponse.Result\x1a\x36\n\x06Result\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x0f\n\x07\x65rr_msg\x18\x03 \x01(\t2z\n\x04Sink\x12\x37\n\x06SinkFn\x12\x14.sink.v1.SinkRequest\x1a\x15.sink.v1.SinkResponse(\x01\x12\x39\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x16.sink.v1.ReadyResponseB8Z6github.com/numaproj/numaflow-go/pkg/apis/proto/sink/v1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "sink_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z6github.com/numaproj/numaflow-go/pkg/apis/proto/sink/v1"
    _globals["_SINKREQUEST"]._serialized_start = 86
    _globals["_SINKREQUEST"]._serialized_end = 235
    _globals["_READYRESPONSE"]._serialized_start = 237
    _globals["_READYRESPONSE"]._serialized_end = 267
    _globals["_SINKRESPONSE"]._serialized_start = 269
    _globals["_SINKRESPONSE"]._serialized_end = 386
    _globals["_SINKRESPONSE_RESULT"]._serialized_start = 332
    _globals["_SINKRESPONSE_RESULT"]._serialized_end = 386
    _globals["_SINK"]._serialized_start = 388
    _globals["_SINK"]._serialized_end = 510
# @@protoc_insertion_point(module_scope)



import logging
import os
import sys

if os.getenv("PYTHONDEBUG"):
    os.environ["PYTHONASYNCIODEBUG"] = "1"


class StdoutFilter(logging.Filter):
    """
    Filter logs with level less than logging.ERROR so they will go to stdout instead
    of default stderr
    """

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno < logging.ERROR


def setup_logging(name):
    formatter = logging.Formatter(
        fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    logger = logging.getLogger(name)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    stdout_handler.addFilter(StdoutFilter())
    stdout_handler.setLevel(logging.INFO)
    logger.addHandler(stdout_handler)

    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setFormatter(formatter)
    stderr_handler.setLevel(logging.ERROR)
    logger.addHandler(stderr_handler)

    return logger



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import sideinput_pb2 as sideinput__pb2


class SideInputStub(object):
    """SideInput is the gRPC service for user-defined Side Inputs.
    It is used to propagate changes in the values of the provided Side Inputs
    which allows access to slow updated data or configuration without needing to retrieve
    it during each message processing.
    Through this service we should should be able to:-
    1) Invoke retrieval request for a single Side Input parameter, which in turn should
    check for updates and return its latest value.
    2) Provide a health check endpoint to indicate whether the service is ready to be used.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RetrieveSideInput = channel.unary_unary(
            "/sideinput.v1.SideInput/RetrieveSideInput",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=sideinput__pb2.SideInputResponse.FromString,
        )
        self.IsReady = channel.unary_unary(
            "/sideinput.v1.SideInput/IsReady",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=sideinput__pb2.ReadyResponse.FromString,
        )


class SideInputServicer(object):
    """SideInput is the gRPC service for user-defined Side Inputs.
    It is used to propagate changes in the values of the provided Side Inputs
    which allows access to slow updated data or configuration without needing to retrieve
    it during each message processing.
    Through this service we should should be able to:-
    1) Invoke retrieval request for a single Side Input parameter, which in turn should
    check for updates and return its latest value.
    2) Provide a health check endpoint to indicate whether the service is ready to be used.
    """

    def RetrieveSideInput(self, request, context):
        """RetrieveSideInput is the endpoint to retrieve the latest value of a given Side Input."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(self, request, context):
        """IsReady is the health check endpoint to indicate whether the service is ready to be used."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_SideInputServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "RetrieveSideInput": grpc.unary_unary_rpc_method_handler(
            servicer.RetrieveSideInput,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=sideinput__pb2.SideInputResponse.SerializeToString,
        ),
        "IsReady": grpc.unary_unary_rpc_method_handler(
            servicer.IsReady,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=sideinput__pb2.ReadyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "sideinput.v1.SideInput", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class SideInput(object):
    """SideInput is the gRPC service for user-defined Side Inputs.
    It is used to propagate changes in the values of the provided Side Inputs
    which allows access to slow updated data or configuration without needing to retrieve
    it during each message processing.
    Through this service we should should be able to:-
    1) Invoke retrieval request for a single Side Input parameter, which in turn should
    check for updates and return its latest value.
    2) Provide a health check endpoint to indicate whether the service is ready to be used.
    """

    @staticmethod
    def RetrieveSideInput(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/sideinput.v1.SideInput/RetrieveSideInput",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            sideinput__pb2.SideInputResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def IsReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/sideinput.v1.SideInput/IsReady",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            sideinput__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from datetime import datetime, timezone, timedelta

MAP_SOCK_PATH = "/var/run/numaflow/map.sock"
MAP_STREAM_SOCK_PATH = "/var/run/numaflow/mapstream.sock"
REDUCE_SOCK_PATH = "/var/run/numaflow/reduce.sock"
SOURCE_TRANSFORMER_SOCK_PATH = "/var/run/numaflow/sourcetransform.sock"
SINK_SOCK_PATH = "/var/run/numaflow/sink.sock"
MULTIPROC_MAP_SOCK_PORT = 55551
MULTIPROC_MAP_SOCK_ADDR = "0.0.0.0"
SIDE_INPUT_SOCK_PATH = "/var/run/numaflow/sideinput.sock"
SOURCE_SOCK_PATH = "/var/run/numaflow/source.sock"

# TODO: need to make sure the DATUM_KEY value is the same as
# https://github.com/numaproj/numaflow-go/blob/main/pkg/function/configs.go#L6
WIN_START_TIME = "x-numaflow-win-start-time"
WIN_END_TIME = "x-numaflow-win-end-time"
MAX_MESSAGE_SIZE = 1024 * 1024 * 64
# TODO: None instead of "EOF" ?
STREAM_EOF = "EOF"
DELIMITER = ":"
DROP = "U+005C__DROP__"
# Watermark are at millisecond granularity, hence we use epoch(0) - 1
# to indicate watermark is not available.
# EVENT_TIME_FOR_DROP is used to indicate that the message is dropped,
# hence excluded from watermark calculation.
EVENT_TIME_FOR_DROP = datetime(1970, 1, 1, tzinfo=timezone.utc) - timedelta(milliseconds=1)
